
// bibhtml: Generates HTML page (with images) from BibTex files
// See: https://www.telecom-paris.fr/~elc/software/
//
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com)
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include <algorithm>
#include <fstream>
#include <sstream>
#include "bib.hpp"
#include "bibentry.hpp"
using namespace std;

static const std::string empty_str = "";

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

BibAuthor::BibAuthor(const std::string& lastname, const std::string& firstname) :
lastname_(lastname),
firstname_(firstname),
sortname_(lastname + ", " + firstname),
authgroup_(nullptr),
stats_(nullptr)
{}

BibGroup::BibGroup(const std::string& name, const std::string& typeName,
                   BibTypes::ID typeID, bool isAuxiliary) :
BibEntry(name, typeName, typeID, isAuxiliary),
name_(name),
stats_(nullptr)
{}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

BibEntry::BibEntry(const std::string& bibKey, const std::string& typeName,
                   BibTypes::ID typeID, bool isAuxiliary) {
  type_ = typeID;
  isAuxiliary_ = isAuxiliary;
  fields_.emplace(BibFields::type, Field{"", typeName});
  fields_.emplace(BibFields::bibkey, Field{"", bibKey});
}

const std::string& BibEntry::bibkey() const {
  return field(BibFields::bibkey);
}

const std::string& BibEntry::field(BibFields::ID id) const {
  if (id <= BibFields::none) return empty_str;
  auto it = fields_.find(id);
  return it == fields_.end() ? empty_str : it->second.parsed;
}

void BibEntry::setField(BibFields::ID id, const std::string& raw, const std::string& parsed) {
  fields_.emplace(id, Field{raw, parsed});
}

void BibEntry::setParsedField(BibFields::ID id, const std::string& parsed) {
  auto it = fields_.find(id);
  if (it == fields_.end()) fields_.emplace(id, Field{"", parsed});
  else it->second.parsed = parsed;  
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void BibEntry::setMediaAndRank(Bib& btw) {
  std::string media;
  switch (type_) {
    case BibTypes::article:
      media = field(BibFields::journal);
      break;
    case BibTypes::inproceedings:
    case BibTypes::proceedings:
    case BibTypes::incollection:
      media = field(BibFields::booktitle);
      break;
    default:
      media = field(BibFields::title);
      break;
  }
  setParsedField(BibFields::media, media);
  
  std::string rank = field(BibFields::rank);
  if (rank.empty() || rank == "?") {
    std::string _media = ccuty::tolower(media);
    const BibEntry* found = nullptr;
    
    for (auto& it : btw.medias()) {
      auto pos = _media.find(it.field(BibFields::title));
      if (pos != std::string::npos) {
        found = &it.entry_;
        rank = it.field(BibFields::rank);
        break;
      }
    }
    if (!found) rank = "?";
    setParsedField(BibFields::rank, rank);
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/* BUGGY
const std::string& BibEntryNode::field(BibFields::ID id) const {
  if (id == BibFields::author || id == BibFields::authgroup) {
    static const std::string empty_str = "";
    if (!author_) return empty_str; else return author_->sortname_;
  }
  else return entry_.field(id);
}
*/

// Return the value of a field (or an empty string if not found).
const std::string& BibEntryNode::field(BibFields::ID id) const {
  return entry_.field(id);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void BibEntries::setMediaAndRank(Bib& btw) {
  for (auto it : *this) it.setMediaAndRank(btw);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

struct Query {
  bool isValid{false};     // true if this query is well formed
  bool hasValues{false};   // true for "field=values" or "field!=values", false for "field"
  bool contains{true};     // true for "field=values", false for "field!=values"
  BibFields::ID field{BibFields::none};
  std::vector<std::string> valueVect;  // vector of values
  std::vector<bool> containsVect;  // containVect[k] = true for field=value, false for field=!value
  
  /// syntax is: field  OR  field=value  OR  field=!value  OR  field=value1|!value2|value3 ...
  /// if field has not value, just find references that have this field
  ///
  Query(std::string query_str, bool abridged) {
    ccuty::strtrim(query_str);
    if (query_str.empty()) return;  // empty queries are discarded
    
    std::string fieldname, values;
    ccuty::strsplit(fieldname, values, query_str, "=");
    
    if (fieldname.empty()) return; // empty field
    hasValues = !values.empty();
    
    if (fieldname.back() != '!') {            // field=...
      if (!hasValues && query_str.back() == '=')
        contains = false;                     // field=""
      else contains = true;                   // field or field=value
    }
    else {                                    // field!=...
      if (!hasValues && query_str.back() == '=')
        contains = true;                      // field!=""
      else contains = false;                  // field! or field!=value
      fieldname.pop_back(); // remove the !
      ccuty::strtrim(fieldname);
      if (fieldname.empty()) return; // empty field
    }
    
    field = BibFields::fieldID(fieldname, abridged);
    if (field <= BibFields::none) {  // unknown field
      isValid = false;
      Bib::error("select",
                 "wrong field '" +fieldname+ "' or invalid syntax (-select:? for help)");
      return;
    }
    else isValid = true;
    
    if (hasValues) {    // if values are specified: "field=value1|!value2|value3"
      ccuty::strsplit(valueVect, values, "|");
      containsVect.resize(valueVect.size());
      
      for (unsigned int k = 0; k < valueVect.size(); ++k) {
        if (valueVect[k][0] != '!')           // field=value
          containsVect[k] = contains;
        else {                                // field=!value
          valueVect[k].erase(0,1);            // remove the !
          containsVect[k] = !contains;
        }
      }
    }
  }
};

struct Queries {
  bool isValid = false;
  std::vector<Query> queries;
  
  /// syntax is: field1=values1 & field2 & field3=values3 ...
  /// if field has no value (eg. field2), just find references that have this field
  ///
  Queries(const std::string& query_str, bool abridged) {
    std::vector<std::string> tokens;
    ccuty::strsplit(tokens, query_str, "&");
    
    for (auto& str : tokens) {
      Query q(str, abridged);
      if (q.isValid) {
        queries.push_back(q);
      }
      else return;
    }
    isValid = true;
  }
};


BibEntries* BibEntries::find(const std::string& query_str,
                             bool ignoreCase, bool abridgedFields) const
{
  Queries queries(ignoreCase ? ccuty::tolower(query_str) : query_str, abridgedFields);
  if (!queries.isValid) return nullptr;

  BibEntries* entries = new BibEntries();
  
  for (auto& it : *this) {   // for all references
    bool ok = true;
    
    for (auto& query : queries.queries) {   // AND on all queries
      std::string refval = it.field(query.field);  // value of the query field
      
      // "field" without values, just verify this ref has/dont has this field
      if (!query.hasValues) {
        if (query.contains && refval.empty()) {ok = false; break;}
        else if (!query.contains && !refval.empty()) {ok = false; break;}
        else continue;
      }
      
      // "field" with values, check if one of the desired values is contained in refval
      else {
        if (ignoreCase) ccuty::lowerize(refval);
        bool match = false;
        
        for (unsigned int k = 0; k < query.valueVect.size(); ++k) {  // OR on valueVect
          if (query.valueVect[k].empty()) {
            if (query.containsVect[k] && !refval.empty()) {match = true; break;}
            else if (!query.containsVect[k] && refval.empty()) {match = true; break;}
            else continue;
          }
          
          else if (refval.empty()) continue;
          
          else {
            if (query.containsVect[k]) {
              if (refval.find(query.valueVect[k]) != std::string::npos) {
                match = true;
                break;
              }
            }
            else {
              if (refval.find(query.valueVect[k]) == std::string::npos) {
                match = true;
                break;
              }
            }
          }
        } // endfor(k)
        if (!match) {ok = false; break;}
      }
      
    } // endfor(query)
    
    if (ok) entries->push_back(it);

  } // endfor(refs)
  return entries;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

struct Compare {
  BibFields::ID id;
  bool reverse;
  Compare(BibFields::ID id, bool reverse) : id(id), reverse(reverse) {}
  
  bool operator() (const BibEntryNode& a, const BibEntryNode& b) {
    const std::string& aValue = a.field(id);
    const std::string& bValue = b.field(id);
    //cerr << "comp " << aValue << " " << bValue << endl;
    return reverse ? (bValue < aValue) : (aValue < bValue);
  }
};

BibEntries* BibEntries::sort(const std::string& pattern, bool abridged) const {
  std::vector<std::string> tokens;
  ccuty::strsplit(tokens, pattern, "&");

  struct FieldRequest {BibFields::ID id; bool reverse;};
  std::vector<FieldRequest> fields;
  bool byauthor = false, byauthgroup = false;
  
  // parse fields
  for (std::string& fieldname : tokens) {
    bool reverse = false;
    if (fieldname.empty()) continue;
    if (fieldname[0] == '-') {  // -field => reverse (ex: -year)
      fieldname.erase(0,1);
      reverse = true;
    }
    BibFields::ID id = BibFields::fieldID(fieldname, abridged);
    if (id <= BibFields::none) {
      Bib::error("sort","wrong field: '" + fieldname + "' (sort:? for help)");
      return nullptr;
    }
    if (id == BibFields::author) byauthor = true;
    // pseudofield BibFields::authgroup
    else if (id == BibFields::authgroup) byauthor = byauthgroup = true;
    fields.push_back({id, reverse});
  }

  // create returned list
  BibEntries* entries  = new BibEntries();
  
  if (byauthor) {
    // duplicate for each author.
    // only consider source entries with index==0 to avoid duplicating duplicates
    for (auto& it : *this) {
      if (it.index_ == 0) {
        int index = 0;
        for (auto& a : it.authors()) {   // duplicate by author
          //if (!byauthgroup || a->authgroup_) { 
            entries->push_back(BibEntryNode(it.entry_, index, a));
            index++;
          //}
        }
      }
    }
  }
  else {
    // copies all refs with index==0 to avoid duplicating duplicates
    for (auto& it : *this) {
      if (it.index_ == 0) {
        entries->push_back(BibEntryNode(it.entry_));
      }
    }
  }

  // sort returned list
  for (auto& it : fields) {
    Compare compare(it.id, it.reverse);
    entries->list::sort(compare);
  }
  return entries;
}
